'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = electronRendererEnhancer;

var _filterObject = require('./utils/filter-object');

var _filterObject2 = _interopRequireDefault(_filterObject);

var _objectMerge = require('./utils/object-merge');

var _objectMerge2 = _interopRequireDefault(_objectMerge);

var _fillShape = require('./utils/fill-shape');

var _fillShape2 = _interopRequireDefault(_fillShape);

var _cloneDeep = require('lodash/cloneDeep');

var _cloneDeep2 = _interopRequireDefault(_cloneDeep);

var _getSubscribeFuncs = require('./getSubscribeFuncs');

var _getSubscribeFuncs2 = _interopRequireDefault(_getSubscribeFuncs);

var _constants = require('./constants');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Creates a store enhancer which allows a redux store to synchronize its data
 * with an electronEnhanced store in the browser process.
 * @param {Object} p - The parameters to the creator
 * @param {Object|Function|true} p.filter - Describes what data should be forwarded to this process
 *                                          from the browser
 * @param {Boolean} p.excludeUnfilteredState - Whether to have all data not passing the filter to be
 *                                             undefined (Helpful to avoid bugs at the cost of
 *                                             performance)
 * @param {Boolean} p.synchronous - Whether dispatches from this process should be processed within
 *                                  this process synchronously, or await an update from the browser process
 * @param {Function} p.postDispatchCallback - A callback to run after a dispatch has occurred.
 * @param {Function} p.preDispatchCallback - A callback to run before an action is dispatched.
 * @param {Function} p.dispatchProxy - In order to allow the actions dispatched by electronEnhancer
 *                                     to pass through the entire store enhancer stack, the final
 *                                     store.dispatch must be injected into redux-electron-store
 *                                     manually.
 * @param {String} p.sourceName - An override to the 'source' property appended to every action
 * @param {Function} p.actionFilter - A function which takes in an action and returns a boolean
 *                                    determining whether to forward the action to other processes
 *                                    or not
 */
function electronRendererEnhancer() {
  var _ref = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  var _ref$filter = _ref.filter;
  var filter = _ref$filter === undefined ? true : _ref$filter;
  var _ref$excludeUnfiltere = _ref.excludeUnfilteredState;
  var excludeUnfilteredState = _ref$excludeUnfiltere === undefined ? false : _ref$excludeUnfiltere;
  var _ref$synchronous = _ref.synchronous;
  var synchronous = _ref$synchronous === undefined ? true : _ref$synchronous;
  var _ref$postDispatchCall = _ref.postDispatchCallback;
  var postDispatchCallback = _ref$postDispatchCall === undefined ? function () {
    return null;
  } : _ref$postDispatchCall;
  var _ref$preDispatchCallb = _ref.preDispatchCallback;
  var preDispatchCallback = _ref$preDispatchCallb === undefined ? function () {
    return null;
  } : _ref$preDispatchCallb;
  var _ref$dispatchProxy = _ref.dispatchProxy;
  var dispatchProxy = _ref$dispatchProxy === undefined ? null : _ref$dispatchProxy;
  var _ref$sourceName = _ref.sourceName;
  var sourceName = _ref$sourceName === undefined ? null : _ref$sourceName;
  var _ref$actionFilter = _ref.actionFilter;
  var actionFilter = _ref$actionFilter === undefined ? function () {
    return true;
  } : _ref$actionFilter;

  return function (storeCreator) {
    return function (reducer, initialState) {
      var _require = require('electron');

      var ipcRenderer = _require.ipcRenderer;
      var remote = _require.remote;

      // If this process is a webview, it will have a guestInstanceId.  Otherwise it is a window

      var rendererId = process.guestInstanceId || remote.getCurrentWindow().id;

      // Get data from the electronEnhanced store in the browser through the global it creates
      var getInitialState = remote.getGlobal(_constants.globalName);
      if (!getInitialState) {
        throw new Error('Could not find electronEnhanced redux store in main process');
      }

      // Prefetch initial state
      var storeData = JSON.parse(getInitialState());
      var preload = excludeUnfilteredState ? (0, _fillShape2.default)(storeData, filter) : storeData;
      var newInitialState = (0, _objectMerge2.default)(initialState || {}, preload);

      var clientId = process.guestInstanceId ? 'webview ' + rendererId : 'window ' + rendererId;
      var currentSource = sourceName || clientId;

      // This flag is toggled to true when events are received
      var mainProcessUpdateFlag = false;

      // Augment the reducer to handle electron enhanced actions that have been forwarded
      // Dispatches from the browser are in the format of {type, data: {updated, deleted}}.
      var parsedReducer = function parsedReducer() {
        var state = arguments.length <= 0 || arguments[0] === undefined ? newInitialState : arguments[0];
        var action = arguments[1];

        if (mainProcessUpdateFlag) {
          mainProcessUpdateFlag = false;
          var data = action.data;
          var filteredState = (0, _filterObject2.default)(state, data.deleted);
          return (0, _objectMerge2.default)(filteredState, data.updated);
        } else {
          var reduced = reducer(state, action);
          return excludeUnfilteredState ? (0, _fillShape2.default)(reduced, filter) : reduced;
        }
      };

      var store = storeCreator(parsedReducer, newInitialState);

      // This must be kept in an object to be accessed by reference
      // by the subscribe function
      var reduxState = { isDispatching: false };

      // Augment the subscribe function to make the listeners happen after the action is forwarded
      var subscribeFuncs = (0, _getSubscribeFuncs2.default)();
      store.subscribe = function (listener) {
        return subscribeFuncs.subscribe(listener, reduxState);
      };

      // Renderers register themselves to the electronEnhanced store in the browser proecss
      ipcRenderer.send(_constants.globalName + '-register-renderer', { filter: filter, clientId: clientId });

      var storeDotDispatch = store.dispatch;
      var doDispatch = function doDispatch(action) {
        reduxState.isDispatching = true;
        try {
          preDispatchCallback(action);
          storeDotDispatch(action);
          postDispatchCallback(action);
        } finally {
          reduxState.isDispatching = false;
        }
      };

      store.dispatch = function (action) {
        if (mainProcessUpdateFlag || !action) {
          doDispatch(action);
        } else {
          action.source = currentSource;

          var shouldForwardUpdate = actionFilter(action);
          if (!shouldForwardUpdate || synchronous) {
            doDispatch(action);
          }

          if (shouldForwardUpdate) {
            ipcRenderer.send(_constants.globalName + '-renderer-dispatch', JSON.stringify({ action: action, clientId: clientId }));
          }
        }

        subscribeFuncs.callListeners();
        return action;
      };

      // Dispatches from other processes are forwarded using this ipc message
      var dispatcher = dispatchProxy || store.dispatch;
      ipcRenderer.on(_constants.globalName + '-browser-dispatch', function (event, _ref2) {
        var action = _ref2.action;
        var sourceClientId = _ref2.sourceClientId;

        var actionParsed = JSON.parse(action);
        if (!synchronous || sourceClientId !== clientId) {
          mainProcessUpdateFlag = true;
          dispatcher(actionParsed);
          subscribeFuncs.callListeners();
        }
      });

      return store;
    };
  };
}